/// <reference types="node" />

import { ClientOptions } from 'ws';
import { EventType } from 'mitt';
import { Handler } from 'mitt';
import { Data as NodeSocketData } from 'ws';
import { WebSocket as WebSocket_2 } from 'ws';
import { WildcardHandler } from 'mitt';

export declare class ClientSocket {
    isNodeEnv: boolean;
    options: Required<ClientSocketOptions>;
    on: {
        <Key extends keyof SocketOnEvent>(type: Key, handler: Handler<SocketOnEvent[Key]>): void;
        (type: "*", handler: WildcardHandler<SocketOnEvent>): void;
    };
    socket: Socket;
    restartNum: number;
    sockSend: SockSend;
    lockReconnect: boolean;
    socketIsClose: boolean;
    socketReadyState: {
        0: string;
        1: string;
        2: string;
        3: string;
    };
    private heartTimer;
    private websocketReconnectTimer;
    constructor(options: ClientSocketOptions);
    createSocket(): Promise<unknown>;
    /**
     * 等待ws的启动 默认等待3次  并不会启动ws 只会等待ws启动成功
     * @params {number} restartMaxNow 该次请求最多重启次数
     * 传递该值的情况为
     * 1.已经实例化并运行过createSocket
     * 2.readyState 不为1（ws不在运行中）
     * 3.需等待重启最多restartMaxNow次
     * **/
    waitCreate(restartMaxNow?: number): Promise<unknown>;
    close(): void;
    private websocketInstance;
    private websocketEvent;
    private websocketReconnect;
    private startHeart;
    private stopHeart;
}

export declare interface ClientSocketOptions {
    url: Url;
    isCacheCmd?: IsCacheCmd;
    cmdAPI?: CmdAPIInit;
    restartMax?: number;
    reconnectTime?: number;
    wsConfig?: ClientOptions;
    wsProtocols?: string;
    heartCmd?: Cmd;
    timeout?: number;
    reqInterceptors?: ReqInterceptors;
    resInterceptors?: ResInterceptors;
}

export declare type Cmd = string;

export declare type CmdAPIInit = Record<Cmd, CmdAPIInitFun>;

export declare type CmdAPIInitFun = (data: CmdAPIItem) => CmdAPIItem;

export declare type CmdAPIItem = {
    uuid: UUID;
    cmd: Cmd;
    content?: ReqContent;
};

export declare type IsCacheCmd = boolean;

export { NodeSocketData }

export declare type ReqContent = Record<string, any>;

export declare type ReqInterceptors = (req: SendReq) => SocketData | Promise<SocketData>;

export declare type ResInterceptors = (res: SocketData | MessageEvent<SocketData>) => SendRes | Promise<SendRes>;

export declare type SendCallback = (data: SendCallbackBack) => any;

export declare interface SendCallbackBack {
    req: SendReq;
    res: SendRes;
    socket: Socket;
}

export declare type SendCmdCache = Record<UUID, SendCmdCacheItem>;

export declare interface SendCmdCacheItem {
    req: CmdAPIItem;
    options: SendOptions;
    resolve: any | PromiseConstructor["resolve"];
    reject: any | PromiseConstructor["reject"];
    status: 'pending' | 'resolved' | 'rejected';
    callback?: SendCallback;
    onTimeout?: NodeJS.Timeout;
}

export declare interface SendOptions {
    isOff?: boolean;
    isCacheCmd?: IsCacheCmd;
    timeout?: number;
    reqInterceptors?: ReqInterceptors;
    resInterceptors?: ResInterceptors;
}

export declare interface SendReq extends CmdAPIItem {
}

export declare interface SendRes extends CmdAPIItem {
    error_msg?: string;
    status?: -1 | 0 | 1;
}

export declare type Socket = WebSocket_2 | WebSocket | undefined;

export declare type SocketData = NodeSocketData | WebSocketData;

export declare type SocketOnEvent = {
    'socket-create-status': any;
    'socket-open': any[];
    'socket-error': any[];
    'socket-close': any[];
    'socket-end': any[];
    'socket-message': any[];
    '*': any;
};

export declare type SocketReadyState = 0 | 1 | 2 | 3;

declare class SockSend {
    options: SockSendOptions;
    sendCmdCache: SendCmdCache;
    on: {
        <Key extends EventType>(type: Key, handler: Handler<Record<EventType, unknown>[Key]>): void;
        (type: "*", handler: WildcardHandler<Record<EventType, unknown>>): void;
    };
    constructor(options: SockSendOptions);
    /**
     * 注册事件 把事件存储起来 client -> serve  向serve通信
     * @params cmd 指令
     * @params data 配置项
     * @params data.uuid? 任务id
     * @params data.content? 数据
     * @params data.options? 配置项
     * @params data.options.isOff 是否要被主动注销
     * @params data.options.isCacheCmd 是否缓存
     * @params data.options.timeout 超时时间
     * @params callback({req, res, socket}) 回调函数
     * */
    send(cmd: Cmd, data?: SockSendData, callback?: SendCallback): {
        uuid: string;
        sendSync: Promise<SendCallbackBack>;
    };
    sendSync(cmd: Cmd, data?: SockSendData, callback?: SendCallback): Promise<SendCallbackBack>;
    restartSend(): void;
    sendError(error_msg: string, uuid?: UUID): void;
    sendErrorUuid(error_msg: string, uuid: UUID): void;
    emit(res: SendRes): void;
    off(uuid?: string): void;
}

export declare interface SockSendData {
    content?: ReqContent;
    options?: SendOptions;
    uuid?: UUID;
}

export declare type SockSendOptions = {
    socket: Socket;
    heartCmd: Cmd;
    isCacheCmd: IsCacheCmd;
    cmdAPI: CmdAPIInit;
    reqInterceptors: ReqInterceptors;
    resInterceptors: ResInterceptors;
};

export declare type typeCall = (news: any, olds: any) => void;

export declare type Url = string;

export declare type UUID = string;

export declare interface WatchInter {
    target: object;
    key: string;
    init: (call: typeCall) => void;
    clear: () => void;
}

export declare type WebSocketData = string | ArrayBufferLike | Blob | ArrayBufferView;

export { }
