import mitt from 'mitt';
import { v1 } from 'uuid';
import WS from 'ws';

/*! *****************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */

function __awaiter(thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
}

const emitter = mitt();
class SockSend {
    constructor(options) {
        this.sendCmdCache = {}; // 缓存的接口
        this.on = emitter.on;
        this.options = options;
    }
    /**
     * 注册事件 把事件存储起来 client -> serve  向serve通信
     * @params cmd 指令
     * @params data 配置项
     * @params data.uuid? 任务id
     * @params data.content? 数据
     * @params data.options? 配置项
     * @params data.options.isOff 是否要被主动注销
     * @params data.options.isCacheCmd 是否缓存
     * @params data.options.timeout 超时时间
     * @params callback({req, res, socket}) 回调函数
     * */
    send(cmd, data, callback) {
        data = data || {};
        const sendOptions = Object.assign({ isOff: false }, data.options);
        const cmdAPIData = this.options.cmdAPI[cmd] ? this.options.cmdAPI[cmd]({
            cmd,
            uuid: data.uuid || v1(),
            content: data.content
        }) : null;
        const uuid = (cmdAPIData === null || cmdAPIData === void 0 ? void 0 : cmdAPIData.uuid) || data.uuid || v1();
        // cmd里获取需传输的数据格式
        const req = { uuid, cmd, content: (cmdAPIData === null || cmdAPIData === void 0 ? void 0 : cmdAPIData.content) || data.content };
        const sendSync = new Promise(((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
            var _a;
            const sendCacheItem = {
                req,
                callback,
                resolve,
                reject,
                options: sendOptions,
                status: 'pending', // 'pending' 请求中 | 'resolved' 已完成 | 'rejected' 已失败
            };
            const wsUrl = ((_a = this.options.socket) === null || _a === void 0 ? void 0 : _a.url) || '';
            // 如果配置项包含超时
            if (sendOptions.timeout) {
                sendCacheItem.onTimeout = setTimeout(() => {
                    this.sendError(`连接超时：${wsUrl}; cmd: ${cmd}; ${sendOptions.timeout}ms`, uuid);
                }, sendOptions.timeout);
            }
            this.sendCmdCache[uuid] = sendCacheItem;
            // 如果有socket
            if (this.options.socket && this.options.socket.readyState === 1) {
                const reqData = yield this.options.reqInterceptors(req);
                this.options.socket.send(reqData);
            }
            else {
                this.sendError(`websocket未创建成功 ${wsUrl}`, uuid);
            }
        })));
        return {
            uuid,
            sendSync,
        };
    }
    // 同步方式：慎用！该方法只会接收一次服务端消息
    sendSync(cmd, data, callback) {
        return this.send(cmd, data, callback).sendSync;
    }
    // 重连时重新调用所有未完成事件
    restartSend() {
        for (let uuid in this.sendCmdCache) {
            const { req, status, options, callback } = this.sendCmdCache[uuid];
            const { cmd, content } = req;
            const { isCacheCmd } = options;
            // 重新发送不是心跳的任务 并且未完成的任务
            if (cmd !== this.options.heartCmd && status !== 'resolved' && isCacheCmd) {
                this.send(cmd, {
                    uuid,
                    content,
                    options
                }, callback);
            }
        }
    }
    // 连接失败
    // 可能是所有接口 或 某个接口
    sendError(error_msg, uuid) {
        if (!uuid) {
            for (let uuid in this.sendCmdCache) {
                this.sendErrorUuid(error_msg, uuid);
            }
        }
        else {
            this.sendErrorUuid(error_msg, uuid);
        }
    }
    // 某个接口发送错误消息
    sendErrorUuid(error_msg, uuid) {
        const { status, req } = this.sendCmdCache[uuid];
        const { cmd } = req;
        // 将正在运行中状态变为失败
        if (status === 'pending') {
            this.sendCmdCache[uuid].status = 'rejected';
        }
        const resData = {
            uuid,
            cmd,
            error_msg,
            content: {},
            status: -1
        };
        this.emit(resData);
    }
    emit(res) {
        const { cmd, uuid, content = {} } = res;
        const sendCmdCacheItem = this.sendCmdCache[uuid];
        if ((sendCmdCacheItem === null || sendCmdCacheItem === void 0 ? void 0 : sendCmdCacheItem.status) === 'pending') {
            sendCmdCacheItem.status = 'resolved'; // 已成功返回
        }
        // 如果有超时 则清除超时定时器
        (sendCmdCacheItem === null || sendCmdCacheItem === void 0 ? void 0 : sendCmdCacheItem.onTimeout) && clearTimeout(sendCmdCacheItem === null || sendCmdCacheItem === void 0 ? void 0 : sendCmdCacheItem.onTimeout);
        const sendCallbackBack = {
            req: sendCmdCacheItem === null || sendCmdCacheItem === void 0 ? void 0 : sendCmdCacheItem.req,
            res,
            socket: this.options.socket
        };
        // 订阅方式 发布cmd接口
        // 服务端主动推送
        // serve -> client
        emitter.emit(cmd, sendCallbackBack);
        // 客户端触发  服务端返回
        // client -> serve -> client
        if (!!sendCmdCacheItem) {
            // 回调方式
            sendCmdCacheItem.callback && sendCmdCacheItem.callback(sendCallbackBack);
            // 同步方式 只会接收一次数据
            sendCmdCacheItem.resolve && sendCmdCacheItem.resolve(sendCallbackBack);
            // 调用一次就注销
            sendCmdCacheItem.options.isOff && this.off(uuid);
        }
    }
    // 通过uuid注销事件 避免数据臃肿
    off(uuid) {
        if (uuid) {
            delete this.sendCmdCache[uuid];
        }
        else {
            this.sendCmdCache = {};
            emitter.off("*");
        }
    }
}

class Watch {
    constructor(target, key) {
        this.target = target;
        this.key = key;
    }
    init(call) {
        let initialValue = this.target[this.key];
        return Object.defineProperty(this.target, this.key, {
            get: function () {
                return initialValue;
            },
            set: function (newValue) {
                call && call(newValue, initialValue);
                initialValue = newValue;
            }
        });
    }
    clear() {
        let value = this.target[this.key];
        delete this.target[this.key];
        this.target[this.key] = value;
    }
}

const emitter$1 = mitt();
class ClientSocket {
    constructor(options) {
        this.isNodeEnv = typeof WebSocket === 'undefined'; // 是否为node环境
        this.options = {
            url: '',
            isCacheCmd: false,
            cmdAPI: {},
            restartMax: 3,
            reconnectTime: 2 * 1000,
            wsConfig: {},
            wsProtocols: '',
            heartCmd: 'heart',
            timeout: 5 * 1000,
            // 请求拦截器
            reqInterceptors: (data) => {
                return Buffer.from(JSON.stringify(data));
            },
            // 响应拦截器
            resInterceptors: (data) => {
                return JSON.parse(data.toString());
            },
        };
        // 主要用来获取ws的生命周期
        this.on = emitter$1.on;
        this.restartNum = 0; // 重启次数
        this.lockReconnect = false; // 正在重连socket
        this.socketIsClose = false; // socket已经关闭
        this.socketReadyState = {
            0: 'websocket连接尚未建立, 无法连接到服务端',
            1: 'websocket连接已建立，可以进行通信',
            2: 'websocket连接正在进行关闭',
            3: 'websocket连接已经关闭或者连接不能打开'
        };
        this.options = Object.assign(Object.assign({}, this.options), options);
        this.options.wsConfig.handshakeTimeout = this.options.timeout;
        this.sockSend = new SockSend({
            socket: this.socket,
            heartCmd: this.options.heartCmd,
            isCacheCmd: this.options.isCacheCmd,
            cmdAPI: this.options.cmdAPI,
            reqInterceptors: this.options.reqInterceptors,
            resInterceptors: this.options.resInterceptors, // 响应拦截器
        });
    }
    // 创建websocket
    createSocket() {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise((resolve) => {
                this.restartNum = 0;
                emitter$1.on('socket-create-status', () => {
                    var _a;
                    resolve((_a = this.socket) === null || _a === void 0 ? void 0 : _a.readyState);
                });
                this.websocketInstance();
            });
        });
    }
    /**
     * 等待ws的启动 默认等待3次  并不会启动ws 只会等待ws启动成功
     * @params {number} restartMaxNow 该次请求最多重启次数
     * 传递该值的情况为
     * 1.已经实例化并运行过createSocket
     * 2.readyState 不为1（ws不在运行中）
     * 3.需等待重启最多restartMaxNow次
     * **/
    waitCreate(restartMaxNow = 3) {
        // 如果传递restartMax 不用在意this.restartMax
        return new Promise((rs) => {
            // 如果已经启动成功
            if (this.socket) {
                if (this.socket.readyState === 1) {
                    // ws 已经启动
                    return rs(this.socket.readyState);
                }
                else if (this.socketIsClose) {
                    //  ws已经完全关闭无法重启
                    return rs(this.socket.readyState);
                }
            }
            else {
                this.createSocket();
            }
            // 当前重启次数
            let restartNum = 0;
            let watchRestartNum = new Watch(this, 'restartNum');
            watchRestartNum.init((newsValue, oldValue) => {
                if (newsValue - 1 === oldValue) {
                    restartNum++;
                    console.log(`等待重启 ${this.options.url} ${restartNum}/${restartMaxNow}`);
                    // 等待 restartMaxNow 次后还未启动成功
                    if (restartNum >= restartMaxNow) {
                        this.socket && rs(this.socket.readyState);
                        watchRestartNum.clear();
                    }
                }
                else {
                    // 启动成功后 restartNum 会变为0
                    this.socket && rs(this.socket.readyState);
                    watchRestartNum.clear();
                }
            });
        });
    }
    close() {
        var _a, _b;
        // 不再重启
        this.socketIsClose = true;
        emitter$1.off('*');
        ((_a = this.socket) === null || _a === void 0 ? void 0 : _a.close) && this.socket.close(1000);
        ((_b = this.sockSend) === null || _b === void 0 ? void 0 : _b.off) && this.sockSend.off();
    }
    // 实例化 websocket
    websocketInstance() {
        try {
            if (this.isNodeEnv) {
                this.socket = new WS(this.options.url, this.options.wsConfig);
            }
            else {
                // 浏览器环境
                this.socket = new WebSocket(this.options.url, this.options.wsProtocols);
            }
            this.websocketEvent();
        }
        catch (e) {
            console.log('catch', e);
            // 出错重连
            this.websocketReconnect();
        }
    }
    // 对socket的事件进行注册
    websocketEvent() {
        const that = this;
        const socketEvent = {
            onopen(...args) {
                emitter$1.emit('socket-open', args);
                // 启动成功
                emitter$1.emit('socket-create-status');
                // 启动成功后重启次数归零
                that.restartNum = 0;
                // 更新sockSend里的socket
                that.sockSend.options.socket = that.socket;
                // 继续运行还未完成的事件
                that.sockSend.restartSend();
                // 创建后开始心跳检测
                that.startHeart();
            },
            onerror(...args) {
                emitter$1.emit('socket-error', args);
                emitter$1.emit('socket-create-status');
                const err = args[0];
                console.error('ws onError', err);
            },
            onclose(...args) {
                emitter$1.emit('socket-close', args);
                emitter$1.emit('socket-create-status');
                const err = args[0];
                console.error('ws onClose', err);
                that.stopHeart();
                that.websocketReconnect(); // 重连
            },
            onend(...args) {
                emitter$1.emit('socket-end', args);
                emitter$1.emit('socket-create-status');
                const err = args[0];
                console.error('ws onEnd', err);
            },
            onmessage(...args) {
                return __awaiter(this, void 0, void 0, function* () {
                    emitter$1.emit('socket-message', args);
                    const data = args[0];
                    //拿到任何消息都说明当前连接是正常的  重新开始心跳检测
                    that.startHeart();
                    const messageData = yield that.options.resInterceptors(data);
                    that.sockSend.emit(messageData);
                });
            }
        };
        if (this.isNodeEnv) {
            const socket = this.socket;
            socket.on('open', socketEvent.onopen);
            socket.on('error', socketEvent.onerror);
            socket.on('close', socketEvent.onclose);
            socket.on('end', socketEvent.onend);
            socket.on('message', socketEvent.onmessage);
        }
        else {
            const socket = this.socket;
            socket.onopen = socketEvent.onopen;
            socket.onerror = socketEvent.onerror;
            socket.onclose = socketEvent.onclose;
            socket.onmessage = socketEvent.onmessage;
        }
    }
    // 对ws重连
    websocketReconnect() {
        const socket = this.socket;
        if (this.lockReconnect) {
            console.log(`正在执行重连, ${this.options.url}`);
            return;
        }
        else if (this.socketIsClose) {
            console.log(`socket已经完全关闭后不再重启, ${this.options.url}`);
            return;
        }
        // 最多重启restartMax次(如果restartMax = 0 不重启)
        else if ((this.restartNum >= this.options.restartMax || this.options.restartMax === 0) && this.options.restartMax >= 0) {
            this.socketIsClose = true;
            const error_msg = `wsReadyState: ${socket.readyState}, ${this.socketReadyState[socket.readyState]}, ${socket.url}`;
            this.sockSend.sendError(error_msg);
            console.log(error_msg);
            // 已经没法重启了
            emitter$1.emit('socket-create-status');
            return;
        }
        this.restartNum++;
        console.log('重启socket次数：', this.restartNum, `最多重启次数${this.options.restartMax === -1 ? '无限制' : this.options.restartMax}`, this.options.url);
        this.lockReconnect = true;
        //没连接上会一直重连，设置延迟避免请求过多
        this.websocketReconnectTimer && clearTimeout(this.websocketReconnectTimer);
        this.websocketReconnectTimer = setTimeout(() => {
            this.lockReconnect = false;
            this.websocketInstance();
        }, this.options.reconnectTime);
    }
    // 开始心跳
    startHeart() {
        this.stopHeart();
        this.heartTimer = setInterval(() => {
            this.sockSend && this.sockSend.send(this.options.heartCmd);
        }, this.options.timeout);
    }
    // 停止心跳
    stopHeart() {
        this.heartTimer && clearInterval(this.heartTimer);
    }
}

export { ClientSocket };
